/** \file storage.c
 * Disk read/write routines.
 * Implemented Disk interface: MMC.
 * 
 * v0.5: added support for more MMC card brands, rewritten code for
 * better clarity.
 *
 * This module reads and writes blocks of disk data to/from a global
 * buffer. The logical place to define these globals is where they are
 * needed most: the module that contains the file system functions.
 * By default that means filesys.c.
 * These globals are needed (see filesys.h):
 * - sectorAddress - contains the absolute
 * disk sector number
 * - diskSect[512] for data read/write (part of diskSect union)
 */


#include "storage.h"
#include "board.h"
#include "console.h"
#include "mmc.h"
#include "buffer.h"
#include "vs10xx.h"



//#define MMCDEBUG


/** Initialize the storage system.
 * - returns 0 when successful 
 * - returns 1 if mmc was not properly initialized
 */

Public unsigned char InitStorage(){

  unsigned char result;

  MOSI_TO_MMC = YES;
  ConsoleWrite ("Init: Storage: MMC/SD in SPI mode.\r");

  result=InitMMC();
  if (result==0x0e){ //ok, no support for seek-before-read
    return 0;
  }

  if (result){ //error resulted in MMC startup
    ConsoleWrite("Can't start MMC: ");
    ConsolePutHex8(result);
    ConsolePutChar(13);
    return 1; //MMC Init Error
  }

  ConsoleWrite("InitStorage ok.\r");
  return 0;

}


/** Read one 512 byte disk sector to extern global char[512] diskSect.
 * Returns 0 when successful, error code 6 otherwise.
 */
Public unsigned char ReadDiskSector(unsigned long sectorN){
  MOSI_TO_MMC = YES;
  //ConsolePutChar('s');
  if (SeekSector(sectorN)) return 0x0f; //seek error
  //ConsolePutChar('S');
  if (ReadPhysicalSector()){
    ConsoleWrite(" Read Error ");
    return 0x10; //read error
  }
  return 0; /* All OK return */
}



unsigned char TransferSectorSDI(unsigned long sectorN){  


  MOSI_TO_MMC = YES;
  Mp3DeselectControl(); //Just in case
  Mp3DeselectData(); //Just in case
  
  if (SeekSector(sectorN)) return 0x0f; //seek error
  // code from mmc.c: unsigned char ReadPhysicalSector(){
  MMCSelect();
  MmcWaitForData();


#if 0
  PerformBlockRead();  

  SPI8Clocks(4); //Send 8*4=32 clocks (4 ff's) to MMC to be nice.
  MMCDeselect();
  SPI8Clocks(4); //Again, give the poor MMC some clocks, it likes them.
  
  Mp3SelectData();
  // BlockTransmit in C
  dataBufPtr = diskSect.raw.buf;
  // Use global pointer and unrolled loop for speed
  while (dataBufPtr < diskSect.raw.buf+512){
    SPIPutCharWithoutWaiting(*dataBufPtr++); 
    SPIPutCharWithoutWaiting(*dataBufPtr++); 
    SPIPutCharWithoutWaiting(*dataBufPtr++); 
    SPIPutCharWithoutWaiting(*dataBufPtr++); 
    SPIPutCharWithoutWaiting(*dataBufPtr++); 
    SPIPutCharWithoutWaiting(*dataBufPtr++); 
    SPIPutCharWithoutWaiting(*dataBufPtr++); 
    SPIPutCharWithoutWaiting(*dataBufPtr++); 
    SPIPutCharWithoutWaiting(*dataBufPtr++); 
    SPIPutCharWithoutWaiting(*dataBufPtr++); 
    SPIPutCharWithoutWaiting(*dataBufPtr++); 
    SPIPutCharWithoutWaiting(*dataBufPtr++); 
    SPIPutCharWithoutWaiting(*dataBufPtr++); 
    SPIPutCharWithoutWaiting(*dataBufPtr++); 
    SPIPutCharWithoutWaiting(*dataBufPtr++); 
    SPIPutCharWithoutWaiting(*dataBufPtr++); 
  }
  Mp3DeselectData();
#else
  // FAST DIRECT TRANSFER VERSION
  
  MOSI_TO_MMC = NO; //force MMC data in to be "1"
  SPIPutCharWithoutWaiting(0xff);
  SPIWait();
  Mp3SelectData();

#if 0
  // C version
  {
    unsigned int i;
    unsigned char c = SPI_RESULT_BYTE;
    for (i=0; i<512; i++){
      SPIPutCharWithoutWaiting(c);
      SPIWait();
      c = SPI_RESULT_BYTE;
    }
  }
#else
  // ASM version





   // BlockRead in Assembler

   _asm
        //asm prologue start
        push    psw
        push    acc
        mov     a,r2
        push    acc
        mov     a,r1
        push    acc
        //prologue end
     
     mov r2, #0 // put 256 to r2
3106$:

3122$:
     
     mov R1, _SPDAT
     mov _SPDAT, R1
     nop
     mov A, _SPSTA
     mov A, _SPSTA
     mov A, _SPSTA
     


     mov R1, _SPDAT
     mov _SPDAT, R1
     nop
     mov A, _SPSTA
     djnz R2, 3122$

3108$:        
        //asm epilogue start
        pop     acc
        mov     r1,a
        pop     acc
        mov     r2,a
        pop     acc
        pop     psw	
        //epilogue end
   _endasm;








#endif

  Mp3DeselectData();
  MOSI_TO_MMC = YES;
  SPI8Clocks(4); //Send 8*4=32 clocks (4 ff's) to MMC to be nice.
  MMCDeselect();
  SPI8Clocks(4); //Again, give the poor MMC some clocks, it likes them.
  
#endif



  
  //RED_LED = LED_OFF;
  
  return 0; //ok return
}


#include "midimap.c"
// Apply midi map translation to diskSect
void ApplyMidiMap(){
  static bit midiProgramChange = 0;
  static xdata unsigned char channel;
  unsigned int i;
  i=0;
  while(i<512){
    if (midiProgramChange){
      unsigned char instrument = diskSect.raw.buf[i];
      unsigned char oldInstrument = instrument;
      instrument = midiMap[instrument];

#if 0
      ConsoleWrite("Channel ");
      ConsolePutUInt(channel+1);
      ConsoleWrite("has instrument ");
      ConsolePutUInt(oldInstrument+1);
      if (instrument != oldInstrument){
	ConsoleWrite("-> ");
	ConsolePutUInt(instrument+1);
      }
      ConsolePutChar(13);
#endif
      diskSect.raw.buf[i] = instrument;      
    }
    midiProgramChange = 0;
    if ((diskSect.raw.buf[i] & 0x80)
	&& ((diskSect.raw.buf[i] & 0xf0) == 0xc0)){      
      midiProgramChange = 1;
      channel = diskSect.raw.buf[i] & 0x0f;
    }
    i++;
  }
}


// Transfer a MIDI sector
unsigned char TransferSectorSCI(unsigned long sectorN){  


  MOSI_TO_MMC = YES;
  Mp3DeselectControl(); //Just in case
  Mp3DeselectData(); //Just in case
  
  if (SeekSector(sectorN)) return 0x0f; //seek error
  
  
  // code from mmc.c: unsigned char ReadPhysicalSector(){
  MMCSelect();
  MmcWaitForData();
  
  //GREEN_LED = LED_ON;
  PerformBlockRead();  
  //GREEN_LED = LED_OFF;
  
  SPI8Clocks(4); //Send 8*4=32 clocks (4 ff's) to MMC to be nice.
  MMCDeselect();
  SPI8Clocks(4); //Again, give the poor MMC some clocks, it likes them.
  
  ApplyMidiMap();
  
  //GREEN_LED = LED_ON;
  Mp3DeselectData(); //Just In Case
  // BlockTransmit in C
  dataBufPtr = diskSect.raw.buf;
  // Use global pointer and unrolled loop for speed
  while (dataBufPtr < diskSect.raw.buf+512){
    Mp3SelectControl();
    SPIPutCharWithoutWaiting(VS_WRITE_COMMAND);SPIWait();
    SPIPutCharWithoutWaiting(SCI_IN0);SPIWait();
    SPIPutCharWithoutWaiting(*dataBufPtr++);SPIWait();
    SPIPutCharWithoutWaiting(*dataBufPtr++);SPIWait(); 
    SPIPutCharWithoutWaiting(*dataBufPtr++);SPIWait(); //SciMultiWrite
    SPIPutCharWithoutWaiting(*dataBufPtr++);SPIWait(); 
    SPIPutCharWithoutWaiting(*dataBufPtr++);SPIWait(); //SciMultiWrite
    SPIPutCharWithoutWaiting(*dataBufPtr++);SPIWait(); 
    SPIPutCharWithoutWaiting(*dataBufPtr++);SPIWait(); //SciMultiWrite
    SPIPutCharWithoutWaiting(*dataBufPtr++);SPIWait(); 
    Mp3DeselectControl();
  }
  //GREEN_LED = LED_OFF;
  
  return 0; //ok return
}






/** Dump disk sector to console in hexdump form */
Public void DumpDiskSector(){
  unsigned int ha,la;
  
  ConsoleWrite("\rDiskBlock ");
  ConsolePutUInt(sectorAddress.l);
  ConsoleWrite(":\r");
  for (ha=0; ha<32; ha++){
    ConsolePutHex16 ((ha<<4));
    ConsoleWrite(": ");
    for (la=0; la<16; la++){
      ConsolePutHex8(diskSect.raw.buf[(ha<<4)+la]);
      ConsolePutChar(' ');
      if (la==7){ 
	ConsolePutChar(' ');
      }
    }
    ConsolePutChar(' ');
    for (la=0; la<16; la++){
      if ((diskSect.raw.buf[(ha<<4)+la]) > 30){
	ConsolePutChar(diskSect.raw.buf[(ha<<4)+la]);
      }else{
	ConsolePutChar('.');
      }
    }
    ConsolePutChar('\r');
  } 
}









/** Write one 512  byte disk sector from extern global diskSect.
 * Sector address (0=first 512B sector, 1=2nd. sector, etc.)
 * is in extern global unsigned long sectorAddress
 * \warning Unimplemented!
 */

Public void WriteDiskSector(unsigned long sectorN){
  MOSI_TO_MMC = YES;
  sectorAddress.l = sectorN;
  dataBufPtr = diskSect.raw.buf;
  WritePhysicalSector();
}

