/** \file vs10xx.c
 * Functions for interfacing with the mp3 player chip.
 * \todo safe rewind
 * \todo VS1003 WMA "wma-bytes-left" variable adjustment at ff/rew
 */

#include "vs10xx.h"
#include "lcd.h"
#include "filesys.h"
#include "storage.h"
#include "ui.h"




/** Read the 16-bit value of a VS10xx register */
unsigned int Mp3ReadRegister (unsigned char addressbyte){
  unsigned int resultvalue = 0;

  Mp3SelectControl();
  SPIPutCharWithoutWaiting(VS_READ_COMMAND);
  SPIPutChar((addressbyte));
  SPIPutChar(0xff);
  SPIWait();
  resultvalue = (SPI_RESULT_BYTE) << 8;
  SPIPutCharWithoutWaiting(0xff);
  SPIWait();
  resultvalue |= (SPI_RESULT_BYTE);
  Mp3DeselectControl();
  return resultvalue;
}
  
bit vs1103_recording_active;
bit vs1103_record_mixer;
bit vs1103_use_line_input;
bit vs1103_do_reset;
bit vs1103_mono_output;
bit vs1103_terminate;
bit vs1103_earSpkHi;
bit vs1103_earSpkLo;
bit vs1103_real_time_midi;

unsigned char currentMicGain;

void LoadPatch();
void StartPatch();

void Vs1103SetMode(){
  unsigned char hiMode, loMode;
  
  hiMode = SM_SDINEW;
  loMode = SM_SDI_ADPCM;

  if (vs1103_use_line_input) hiMode |= SM_LINE_IN;
  if (vs1103_recording_active) hiMode |= SM_RECORD_ADPCM;
  if (vs1103_earSpkLo) hiMode |= SM_EARSPK_1;
  if (vs1103_earSpkHi) hiMode |= SM_EARSPK_2;
  
  if (vs1103_do_reset) loMode |= SM_RESET;
  if (vs1103_mono_output) loMode |= SM_DIFF;
  if (vs1103_record_mixer) loMode |= SM_RECORD_MIXER;
  if (vs1103_terminate) loMode |= SM_OUTOFMIDI;
  if (vs1103_real_time_midi) loMode |= SM_RT_MIDI;

  if (vs1103_real_time_midi){
    //Enable VS1103 UART interrupt
    Mp3WriteRegister(SCI_WRAMADDR, 0xc0, 0x1a);
    Mp3WriteRegister(SCI_WRAM, 0x00, 0x3f);
  }

  Mp3WriteRegister (SCI_MODE, hiMode, loMode);
  ConsoleWrite("SetMode: ");
  ConsolePutHex8(hiMode);
  ConsolePutHex8(loMode);
  ConsolePutChar(13);

  
  while (!MP3_DREQ)
    ;//wait for operation to complete


  if (!vs1103_real_time_midi){
    //When real-time midi input is not active,
    //Uart RX goes to monitor interrupt. To prevent
    //from falling to the monitor, disable RX interrupt
    //now
    Mp3WriteRegister(SCI_WRAMADDR, 0xc0, 0x1a);
    Mp3WriteRegister(SCI_WRAM, 0x00, 0x0f);
  } 
  
  StartPatch();
  
}


/** Soft Reset of VS10xx (Between songs) */
void Mp3SoftReset(){

  /* Soft Reset of VS10xx */
  vs1103_do_reset = 1;
  Vs1103SetMode();
  vs1103_do_reset = 0;
 
  Delay(1); /* One millisecond delay */
  while (!MP3_DREQ) /* wait for startup */
    ;
  
  Mp3SetMixerVolume(25,25,25);

  Delay(1);
  //Mp3WriteRegister(SCI_CLOCKF, 0xa6, 0x96);


  // Set default microphone gain
  Delay(1);
  Mp3WriteRegister(SCI_ADPCMRECCTL, USE_512_BYTE_DREQ, 0x00 | 20);
  
  
  Delay(1); /* One millisecond delay */
  
  /* Send null bytes to data interface */
  Mp3SelectData();
  SPIPutCharWithoutWaiting(0);
  SPIPutChar(0);
  SPIPutChar(0);
  SPIPutChar(0);
  SPIWait();
  Mp3DeselectData();

}




/** Reset VS10xx */
void Mp3Reset(){

  vs1103_recording_active = 0;
  vs1103_record_mixer = 1;
  vs1103_use_line_input = 1;
  vs1103_do_reset = 0;
  vs1103_mono_output = 0;
  vs1103_terminate = 0;
  vs1103_earSpkHi = 0;
  vs1103_earSpkLo = 0;
  vs1103_real_time_midi = 0;
  
  SPISetSlowClock();

  Mp3PutInReset();
  Delay(1);

  /* Send dummy SPI byte to initialize SPI */
  SPIPutCharWithoutWaiting(0xFF);

  /* Un-reset MP3 chip */
  Mp3DeselectControl();
  Mp3DeselectData();
  Mp3ReleaseFromReset();

  Delay(1);
  while(!MP3_DREQ);
  /* Start up analog with no sound */
  Mp3SetVolume(0xfe,0xfe);
  Mp3SetMixerVolume(0,0,0);
  LoadPatch();
  Vs1103SetMode();

  /* Set clock register, doubler etc. */
  Mp3WriteRegister(SCI_CLOCKF, 0xa6, 0x96); //oli 0xa66c


  Delay(500);

  Mp3SetMixerVolume(1,1,1);
  //ConsoleWrite("Mixer111 ");
  //Delay(1000);
  //Mp3SetVolume(0xfe,0xfe);
  //ConsoleWrite("Mixer 1,1,1 SecondDelay");
  //Delay(2000);
  Mp3SetVolume(volume,volume);

  SPISetFastClock();

  Mp3SoftReset();

  //ConsoleWrite("Clock:");ConsolePutHex16(Mp3ReadRegister(SCI_CLOCKF));
  ConsoleWrite ("\rInit: VS10XX\r");

}


/** VS10xx Sine Test Function - Good getting started example */ 
void VsSineTest(){

  /* Reset MP3 chip */
  Mp3PutInReset();       /* Pull xRESET low -> hardware reset */
  Delay(100);            /* 100 ms delay */

  /* Send dummy SPI byte to initialize SPI of Atmel microcontroller */
  SPIPutCharWithoutWaiting(0xFF);

  /* Un-reset MP3 chip */
  Mp3DeselectControl();  /* Pull xCS high    */
  Mp3DeselectData();     /* Pull xDCS high   */
  Mp3ReleaseFromReset(); /* Pull xRESET high */
  Delay(100);            /* 100 ms delay     */

  //GREEN_LED = LED_ON;
  //RED_LED = LED_ON;

  /* VS10xx Application Notes, chapter 4.8 ---------------------------------*/
  /* As an example, let's write value 0x0820 to register 00 byte by byte    */
  Mp3SelectControl();    /* Now SPI writes go to SCI port                   */
  SPIPutChar(0x02);      /* Send SPI Byte, then wait for byte to be sent.   */
  SPIPutChar(0x00);      /* 0x02 was WRITE command, 0x00 is register number */
  SPIPutChar(0x08);      /* This byte goes to MSB                           */
  SPIPutChar(0x20);      /* ..and this is LSB. (0x20=Allow Test Mode)       */
  SPIWait();             /* Wait until Atmel MCU signals SPI write complete */
  Mp3DeselectControl();  /* Now SPI writes don't go to SCI port             */

  while (!MP3_DREQ)      /* Wait for DREQ = 1                               */
    ; 			 /* Do nothing while waiting for DREQ = 1           */

  /* Send a Sine Test Header to Data port                                   */
  Mp3SelectData();       /* Now SPI writes go to SDI port                   */

  SPIPutChar(0x53);      /* - This is a special VLSI Solution test header - */
  SPIPutChar(0xef);      /* - that starts a sine sound. It's good for     - */
  SPIPutChar(0x6e);      /* - testing your code, the chip and also for    - */
  SPIPutChar(0x44);      /* - seeing if your MP3 decoder was manufactured - */
  SPIPutChar(0x00);      /* - by VLSI Solution oy. ------------------------ */
  SPIPutChar(0x00);
  SPIPutChar(0x00);
  SPIPutChar(0x00);
  SPIWait();
  Mp3DeselectData();
  
  //RED_LED = LED_OFF;
  Delay (500);           /* 500 ms delay */
  //GREEN_LED = LED_OFF;

  /* Stop the sine test sound */
  Mp3SelectData();
  SPIPutChar(0x45);
  SPIPutChar(0x78);
  SPIPutChar(0x69);
  SPIPutChar(0x74);
  SPIPutChar(0x00);
  SPIPutChar(0x00);
  SPIPutChar(0x00);
  SPIPutChar(0x00);
  SPIWait();
  Mp3DeselectData();

  Delay(500);            /* 500 ms delay */
}  

/** Send 2048 zeros. */
void SendZerosToVS10xx(){
  Mp3SelectData();
  SPIPutCharWithoutWaiting(0);
  for (temp.i=0; temp.i<2048; temp.i++){ /* TESTING 1048 TESTING */
    while (!MP3_DREQ)
      ;
    SPIPutChar(0);
  }
  SPIWait();
  Mp3DeselectData();
}  


/** Send a number of disk sectors to vs10xx.
 * Starting from current value in global variable sectorAddress,
 * sends a number of disk sectors to vs10xx and returns. */
unsigned char PlayDiskSectors (unsigned int nSectorsToPlay){

  while (nSectorsToPlay){

    if (MP3_DREQ){
      TransferSectorSDI(sectorAddress.l);
      sectorAddress.l++;
      nSectorsToPlay--;
    }

    AvailableProcessorTime(); //Call the User Interface handler

    /* If playing state is something else than "play normally",
       exit returning the request number. */
    if ((playingState==PS_END_OF_SONG)||
	(playingState==PS_NEXT_SONG)||
	(playingState==PS_RECORDING)||
	(playingState==PS_PREVIOUS_SONG)){
      return playingState;
    }
    
  }
  return 0; //OK Exit
}



void Mp3SetMixerVolume(unsigned char gain1, 
		       unsigned char gain2, 
		       unsigned char gainInput){
  unsigned int i;

  i=0x8000; // Mixer Mode On
  i |= (gainInput << 10);
  i |= (gain2 << 5);
  i |= (gain1);

  Mp3WriteRegister(SCI_MIXERVOL, (i >> 8), (i & 0xff));

}

void Vs1103B_SetMidiTranspose(int transpose){
  Mp3WriteRegister(SCI_WRAMADDR, 0x11, 0x32);
  while (!MP3_DREQ);
  Mp3WriteRegister(SCI_WRAM,(transpose >> 8),(transpose));
}

void Vs1103B_SetChannelMask(unsigned int newMask){
  Mp3WriteRegister(SCI_WRAMADDR, 0x17, 0x15);
  while (!MP3_DREQ);
  Mp3WriteRegister(SCI_WRAM,(newMask >> 8),(newMask));
}

void Vs1103SetMicGain(unsigned char gain){
  if (gain>31) gain=31;
  currentMicGain = gain;
  Mp3WriteRegister(SCI_ADPCMRECCTL, USE_512_BYTE_DREQ, 0x40 | gain);
}

void Vs1103CancelWave(){
  Mp3WriteRegister(SCI_ADPCMRECCTL, USE_512_BYTE_DREQ, 0x80 | currentMicGain);

  while (Mp3ReadRegister(SCI_ADPCMRECCTL) & 0x0080)
    ;
}


void Vs1103CancelMidi(){
  vs1103_terminate = 1;
  Vs1103SetMode();	  
  vs1103_terminate = 0;
  // The below lines are recommended but take a little
  // too much time for our non-multitasking 8051 
  // architecture:
  //while (Mp3ReadRegister(SCI_MODE) & SM_OUTOFMIDI)
  //  ;

}
